/*
Software for the nanobot pcb
(C) 2014 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <ftdi.h>
#include <math.h>

struct ftdi_context ftdi;
unsigned char sendbuff[10240*1024];
int sendpos;

/*
This is software for the MosFET-powered version. It expects one 16-bit shift
register driver connected to the FT232. 8 MosFETs are connected to the upper
8 outputs driving the X direction, 8 are connected to the lower 8 outputs
to drive the Y direction.
*/

/*
Pinout for the PCB:
1 - ground
2 - /latch
3 - clock
4 - data
5 - Vdd,logic
6 - Vdd,magn
*/


//Masks for the pin signals
#define NLAT 0x10 //DTR, pin 2
#define CLK 0x02 //RX, pin 5
#define SIN 0x01 //TX, pin 1
#define NC  0x08 //CTS, pin 11


//Error handling routine.
int error(const char *err) {
	printf("Error at %s: %s\n", err, ftdi_get_error_string(&ftdi));
	exit(0); //Dirty. Disable to continue after errors.
	return 0;
}

void flush() {
		if (ftdi_write_data(&ftdi, sendbuff, sendpos)<0) error("writing buffer");
		sendpos=0;
}

void magnetizeReal(int x, int y) {
	int i, b;
	static int n=0;

	printf("x: % 3d   y:% 3d\n", x,y);

	for (i=0; i<8; i++) {
		b=SIN;
		if (x==i) b=0;
		sendbuff[sendpos++]=b;
		sendbuff[sendpos++]=b|CLK;
		sendbuff[sendpos++]=b;
	}

	for (i=0; i<8; i++) {
		b=SIN;
		if (y==i) b=0;
		sendbuff[sendpos++]=b;
		sendbuff[sendpos++]=b|CLK;
		sendbuff[sendpos++]=b;
	}

	sendbuff[sendpos++]=NLAT;
	sendbuff[sendpos++]=0;

	n++;
	if (n==16) {
		printf("x: % 3d   y:% 3d\n", x,y);
		flush();
		n=0;
	}
}


//This routine will use error diffusion to dither a position that's not an integer, basically
//microstepping the magnet.
void magnetize(double x, double y) {
	int rx, ry;
	static double errx=0, erry=0;
	if (x>=0) x+=errx;
	if (y>=0) y+=erry;
	rx=(x+0.5);
	ry=(y+0.5);
	magnetizeReal(rx,ry);
	if (x>=0) errx=x-rx;
	if (y>=0) erry=y-ry;
}


//Main function
int main(int argc, char **argv) {
	double t=0;
	int x;
	int n=0;

	//Open FTDI communications
	if (ftdi_init(&ftdi)<0) return error("init");
	if (ftdi_usb_open(&ftdi, 0x0403, 0x6001)<0) return error("open");
	if (ftdi_set_bitmode(&ftdi, SIN|CLK|NLAT, BITMODE_BITBANG)<0) error("bitmode");
	if (ftdi_set_baudrate(&ftdi, 3000000/4)<0) error("baudrate");
	ftdi_usb_purge_buffers(&ftdi);
	sendpos=0;

	//Make the magnet move clockwise in large steps
	for (n=0; n<50; n++) {
		for (x=0; x<200; x++) {
			magnetize(sin(t)*3.5+4, cos(t)*3.5+4);
		}
		magnetizeReal(-1, -1);
		flush();
		usleep(100000);
		t+=0.3;
	}

	sleep(1);

	//Make the magnet move counter-clockwise, in tinier steps.
	for (n=0; n<200; n++) {
		for (x=0; x<100; x++) {
			magnetize(-sin(t)*3.5+4, cos(t)*3.5+4);
		}
		magnetizeReal(-1, -1);
		flush();
		usleep(10000);
		t+=0.15;
	}

}
